@@ -97,6 +97,8 @@ struct _EBookBackendMAPIGALPrivate
 	mapi_id_t fid;
 	int mode;
 	gboolean marked_for_offline;
+	GThread *build_cache_thread;
+	gboolean kill_cache_build;
 	gboolean is_cache_ready;
 	gboolean is_summary_ready;
 	gboolean is_writable;
@@ -152,7 +154,7 @@ build_cache (EBookBackendMAPIGAL *ebmapi)
 	e_book_backend_summary_add_contact (priv->summary, contact);		
 	g_object_unref(contact);
 
-	for (i=0;i<contacts_array->len;i++) {
+	for (i = 0; i < contacts_array->len && !priv->kill_cache_build; i++) {
 		EContact *contact = e_contact_new ();
 		ExchangeMAPIGALEntry *gal_entry = contacts_array->pdata[i];
 		char *uid;
@@ -169,16 +171,16 @@ build_cache (EBookBackendMAPIGAL *ebmapi)
 		g_free (uid);
 	}
 
-	g_ptr_array_free(contacts_array, TRUE);
+	g_ptr_array_free (contacts_array, TRUE);
 
-	tmp = g_strdup_printf("%d", (int)time (NULL));
+	tmp = g_strdup_printf("%d", priv->kill_cache_build ? 0 : (int)time (NULL));
 	e_book_backend_cache_set_time (priv->cache, tmp);
 	printf("setting time  %s\n", tmp);
 	g_free (tmp);
 	e_file_cache_thaw_changes (E_FILE_CACHE (priv->cache));
 	e_book_backend_summary_save (priv->summary);
-	priv->is_cache_ready = TRUE;
-	priv->is_summary_ready = TRUE;
+	priv->is_cache_ready = !priv->kill_cache_build;
+	priv->is_summary_ready = !priv->kill_cache_build;
 	return NULL;
 }
 
@@ -234,7 +236,7 @@ e_book_backend_mapi_gal_authenticate_user (EBookBackend *backend,
 		
 	case GNOME_Evolution_Addressbook_MODE_REMOTE:
 		
-		if (!exchange_mapi_connection_new (priv->profile, NULL))
+		if (!exchange_mapi_connection_new (priv->profile, passwd))
 			return e_data_book_respond_authenticate_user (book, opid,GNOME_Evolution_Addressbook_OtherError);
 
 		if (priv->cache && priv->is_cache_ready) {
@@ -242,9 +244,12 @@ e_book_backend_mapi_gal_authenticate_user (EBookBackend *backend,
 //			g_thread_create ((GThreadFunc) update_cache, 
 	//					  backend, FALSE, backend);
 		} else if (priv->marked_for_offline && !priv->is_cache_ready) {
-			/* Means we dont have a cache. Lets build that first */
-			printf("Preparing to build cache\n");
-			g_thread_create ((GThreadFunc) build_cache, backend, FALSE, NULL);
+			if (!priv->build_cache_thread) {
+				/* Means we dont have a cache. Lets build that first */
+				printf("Preparing to build cache\n");
+				priv->kill_cache_build = FALSE;
+				priv->build_cache_thread = g_thread_create ((GThreadFunc) build_cache, backend, TRUE, NULL);
+			}
 		} 
 		e_book_backend_set_is_writable (backend, FALSE);
 		e_data_book_respond_authenticate_user (book, opid, GNOME_Evolution_Addressbook_Success);
@@ -362,9 +367,14 @@ e_book_backend_mapi_gal_load_source (EBookBackend *backend,
 
 		/* Load the cache as well.*/
 		if (e_book_backend_cache_exists (priv->uri)) {
+			gchar *last_time;
+
 			printf("Loading the cache\n");
 			priv->cache = e_book_backend_cache_new (priv->uri);
-			priv->is_cache_ready = TRUE;
+
+			last_time = e_book_backend_cache_get_time (priv->cache);
+			priv->is_cache_ready = last_time && !g_str_equal (last_time, "0");
+			g_free (last_time);
 		}
 		//FIXME: We may have to do a time based reload. Or deltas should upload.
 	} else {
@@ -408,7 +418,7 @@ e_book_backend_mapi_gal_load_source (EBookBackend *backend,
 }
 
 static void 
-e_book_backend_mapi_gal_set_mode (EBookBackend *backend, int mode)
+e_book_backend_mapi_gal_set_mode (EBookBackend *backend, GNOME_Evolution_Addressbook_BookMode mode)
 {
 	EBookBackendMAPIGALPrivate *priv = ((EBookBackendMAPIGAL *) backend)->priv;
 
@@ -446,7 +456,12 @@ e_book_backend_mapi_gal_dispose (GObject *object)
 		g_free (priv->uri);
 		priv->uri = NULL;
 	}
-	
+
+	if (priv->build_cache_thread) {
+		priv->kill_cache_build = TRUE;
+		g_thread_join (priv->build_cache_thread);
+		priv->build_cache_thread = NULL;
+	}
 }
 
 typedef struct {
@@ -786,13 +801,25 @@ init_closure (EDataBookView *book_view, EBookBackendMAPIGAL *bg)
 	return closure;
 }
 
+static void
+e_book_backend_mapi_gal_get_contact (EBookBackend *backend,
+				     EDataBook    *book,
+				     guint32       opid,
+				     const char   *id)
+{
+	if (enable_debug)
+		printf ("mapi: get contact %s\n", id);
+
+	e_data_book_respond_get_contact (book, opid, GNOME_Evolution_Addressbook_RepositoryOffline, NULL);
+}
+
 static void
 e_book_backend_mapi_gal_get_contact_list (EBookBackend *backend,
 					    EDataBook    *book,
 					    guint32       opid,
 					    const char   *query )
 {
-	EBookBackendMAPIGALPrivate *priv = ((EBookBackendMAPIGAL *) backend)->priv;
+	/*EBookBackendMAPIGALPrivate *priv = ((EBookBackendMAPIGAL *) backend)->priv;*/
 
 	if(enable_debug)
 		printf("mapi: get contact list %s\n", query);
@@ -803,7 +830,6 @@ e_book_backend_mapi_gal_get_contact_list (EBookBackend *backend,
 	return;
 }
 
-
 static void
 e_book_backend_mapi_gal_start_book_view (EBookBackend  *backend,
 					   EDataBookView *book_view)
@@ -827,6 +853,42 @@ e_book_backend_mapi_gal_stop_book_view (EBookBackend  *backend,
 	/* FIXME : provide implmentation */
 }
 
+static void
+e_book_backend_mapi_gal_get_changes (EBookBackend *backend, EDataBook *book, guint32 opid, const char *change_id)
+{
+	if (enable_debug)
+		printf ("mapi: get changes\n");
+
+	e_data_book_respond_get_changes (book, opid, GNOME_Evolution_Addressbook_RepositoryOffline, NULL);
+}
+
+static void 
+e_book_backend_mapi_gal_get_supported_auth_methods (EBookBackend *backend, EDataBook *book, guint32 opid)
+{
+	GList *auth_methods = NULL;
+	char *auth_method;
+	
+	if (enable_debug)
+		printf ("mapi get_supported_auth_methods...\n");
+
+	auth_method =  g_strdup_printf ("plain/password");
+	auth_methods = g_list_append (auth_methods, auth_method);
+	e_data_book_respond_get_supported_auth_methods (book,
+							opid,
+							GNOME_Evolution_Addressbook_Success,
+							auth_methods);  
+	g_free (auth_method);
+	g_list_free (auth_methods);	
+}
+
+static GNOME_Evolution_Addressbook_CallStatus
+e_book_backend_mapi_gal_cancel_operation (EBookBackend *backend, EDataBook *book)
+{
+	if (enable_debug)
+		printf ("mapi cancel_operation...\n");
+	return GNOME_Evolution_Addressbook_CouldNotCancel;
+}
+
 static void
 e_book_backend_mapi_gal_remove (EBookBackend *backend, EDataBook *book, guint32 opid)
 {
@@ -850,12 +912,16 @@ static void e_book_backend_mapi_gal_class_init (EBookBackendMAPIGALClass *klass)
 	parent_class->load_source		 = e_book_backend_mapi_gal_load_source;
 	parent_class->get_static_capabilities    = e_book_backend_mapi_gal_get_static_capabilities;
 
+	parent_class->get_contact                = e_book_backend_mapi_gal_get_contact;
 	parent_class->get_contact_list           = e_book_backend_mapi_gal_get_contact_list;
 	parent_class->start_book_view            = e_book_backend_mapi_gal_start_book_view;
 	parent_class->stop_book_view             = e_book_backend_mapi_gal_stop_book_view;
+	parent_class->get_changes                = e_book_backend_mapi_gal_get_changes;
 	parent_class->authenticate_user          = e_book_backend_mapi_gal_authenticate_user;
 	parent_class->get_supported_fields	 = e_book_backend_mapi_gal_get_supported_fields;
 	parent_class->get_required_fields	 = e_book_backend_mapi_gal_get_required_fields;
+	parent_class->get_supported_auth_methods = e_book_backend_mapi_gal_get_supported_auth_methods;
+	parent_class->cancel_operation		 = e_book_backend_mapi_gal_cancel_operation;
 	parent_class->set_mode                   = e_book_backend_mapi_gal_set_mode;
 	parent_class->remove			 = e_book_backend_mapi_gal_remove;
 	object_class->dispose                    = e_book_backend_mapi_gal_dispose;
@@ -893,6 +959,8 @@ static void	e_book_backend_mapi_gal_init (EBookBackendMAPIGAL *backend)
 	/* Priv Struct init */
 	backend->priv = priv;
 
+	priv->build_cache_thread = NULL;
+
 /*	priv->marked_for_offline = FALSE;
 	priv->uri = NULL;
 	priv->cache = NULL;
