@@ -79,12 +79,6 @@ typedef struct {
 } flags_diff_t;
 
 /*For collecting summary info from server*/
-typedef struct {
-	GSList *items_list;
-	GTimeVal last_modification_time;
-	CamelFolder *folder;
-	CamelFolderChangeInfo *changes;
-}fetch_items_data;
 
 static CamelMimeMessage *mapi_folder_item_to_msg( CamelFolder *folder, MapiItem *item, CamelException *ex );
 static void mapi_update_cache (CamelFolder *folder, GSList *list, CamelFolderChangeInfo **changeinfo,
@@ -300,7 +294,7 @@ read_item_common (MapiItem *item, uint32_t ulPropTag, gconstpointer prop_data)
 }
 
 static gboolean
-fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
+fetch_items_summary_cb (FetchItemsCallbackData *item_data, gpointer data)
 {
 	fetch_items_data *fi_data = (fetch_items_data *)data;
 	
@@ -380,7 +374,8 @@ fetch_items_cb (FetchItemsCallbackData *item_data, gpointer data)
 	*slist = g_slist_prepend (*slist, item);
 
 	/*Write summary to db in batches of SUMMARY_FETCH_BATCH_COUNT items.*/ 
-	if ( (item_data->index % SUMMARY_FETCH_BATCH_COUNT == 0) || item_data->index == item_data->total-1) {
+	if ((item_data->index % SUMMARY_FETCH_BATCH_COUNT == 0) ||
+	     item_data->index == item_data->total-1) {
 		mapi_update_cache (fi_data->folder, *slist, &fi_data->changes, NULL, false);
 		g_slist_foreach (*slist, (GFunc)mapi_item_free, NULL);
 		g_slist_free (*slist);
@@ -787,7 +782,7 @@ mapi_sync_deleted (CamelSession *session, CamelSessionThreadMsg *msg)
 
 	camel_operation_end (NULL);
 
-	camel_object_trigger_event (m->folder, "folder_changed", changes);
+	/* camel_object_trigger_event (m->folder, "folder_changed", changes); */
 	camel_folder_change_info_free (changes);
 
 	/* Discard server uid list */
@@ -966,6 +961,57 @@ mapi_sync (CamelFolder *folder, gboolean expunge, CamelException *ex)
 	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 }
 
+gboolean
+camel_mapi_folder_fetch_summary (CamelStore *store, const mapi_id_t fid, struct mapi_SRestriction *res,
+				 struct SSortOrderSet *sort, fetch_items_data *fetch_data, guint32 options)
+{
+	gboolean status;
+	CamelMapiStore *mapi_store = (CamelMapiStore *) store;
+
+	const guint32 summary_prop_list[] = {
+		PR_INTERNET_CPID,
+		PR_SUBJECT,
+		PR_SUBJECT_UNICODE,
+		PR_MESSAGE_SIZE,
+		PR_MESSAGE_DELIVERY_TIME,
+		PR_MESSAGE_FLAGS,
+		PR_SENT_REPRESENTING_NAME,
+		PR_SENT_REPRESENTING_NAME_UNICODE,
+		PR_SENT_REPRESENTING_EMAIL_ADDRESS,
+		PR_SENT_REPRESENTING_EMAIL_ADDRESS_UNICODE,
+		PR_SENT_REPRESENTING_ADDRTYPE,
+		PR_SENT_REPRESENTING_ADDRTYPE_UNICODE,
+		PR_LAST_MODIFICATION_TIME,
+		PR_INTERNET_MESSAGE_ID,
+		PR_INTERNET_REFERENCES,
+		PR_IN_REPLY_TO_ID,
+		PR_DISPLAY_TO,
+		PR_DISPLAY_TO_UNICODE,
+		PR_DISPLAY_CC,
+		PR_DISPLAY_CC_UNICODE,
+		PR_DISPLAY_BCC,
+		PR_DISPLAY_BCC_UNICODE,
+		PR_TRANSPORT_MESSAGE_HEADERS,
+		PR_TRANSPORT_MESSAGE_HEADERS_UNICODE
+	};
+
+	/*TODO : Check for online state*/
+
+	camel_operation_start (NULL, _("Fetching summary information for new messages in")); /* %s"), folder->name); */
+
+	CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
+
+	status = exchange_mapi_connection_fetch_items  (fid, res, sort, summary_prop_list,
+							G_N_ELEMENTS (summary_prop_list), 
+							NULL, NULL, fetch_items_summary_cb,
+							fetch_data, options);
+
+	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
+
+	camel_operation_end (NULL);
+
+	return status;
+}
 
 void
 mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
@@ -987,27 +1033,6 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 
 	const gchar *folder_id = NULL;
 
-	const guint32 summary_prop_list[] = {
-		PR_INTERNET_CPID,
-		PR_SUBJECT,
-		PR_SUBJECT_UNICODE,
-		PR_MESSAGE_SIZE,
-		PR_MESSAGE_DELIVERY_TIME,
-		PR_MESSAGE_FLAGS,
-		PR_SENT_REPRESENTING_NAME,
-		PR_SENT_REPRESENTING_EMAIL_ADDRESS,
-		PR_SENT_REPRESENTING_ADDRTYPE,
-		PR_LAST_MODIFICATION_TIME,
-		PR_INTERNET_MESSAGE_ID,
-		PR_INTERNET_REFERENCES,
-		PR_IN_REPLY_TO_ID,
-		PR_DISPLAY_TO,
-		PR_DISPLAY_CC,
-		PR_DISPLAY_BCC,
-		PR_TRANSPORT_MESSAGE_HEADERS,
-		PR_TRANSPORT_MESSAGE_HEADERS_UNICODE
-	};
-
 	if (((CamelOfflineStore *) mapi_store)->state == CAMEL_OFFLINE_STORE_NETWORK_UNAVAIL)
 		goto end1;
 
@@ -1083,14 +1108,8 @@ mapi_refresh_folder(CamelFolder *folder, CamelException *ex)
 		if (((CamelMapiFolder *)folder)->type & CAMEL_MAPI_FOLDER_PUBLIC)
 			options |= MAPI_OPTIONS_USE_PFSTORE;
 
-		camel_operation_start (NULL, _("Fetching summary information for new messages in %s"), folder->name);
-
-		status = exchange_mapi_connection_fetch_items  (temp_folder_id, res, sort,
-								summary_prop_list, G_N_ELEMENTS (summary_prop_list), 
-								NULL, NULL, 
-								fetch_items_cb, fetch_data, 
-								options);
-		camel_operation_end (NULL);
+		status = camel_mapi_folder_fetch_summary ((CamelStore *)mapi_store, temp_folder_id, res, sort,
+							  fetch_data, options);
 
 		if (!status) {
 			camel_exception_set (ex, CAMEL_EXCEPTION_SERVICE_INVALID, _("Fetching items failed"));
@@ -1134,6 +1153,7 @@ static const uint32_t camel_GetPropsList[] = {
 	PR_MID, 
 	PR_INTERNET_CPID,
 
+
 	PR_TRANSPORT_MESSAGE_HEADERS,
 	PR_TRANSPORT_MESSAGE_HEADERS_UNICODE,
 	PR_MESSAGE_CLASS, 
@@ -1729,11 +1749,13 @@ mapi_folder_get_message( CamelFolder *folder, const char *uid, CamelException *e
 		options |= MAPI_OPTIONS_USE_PFSTORE;
 	} 
 
+	CAMEL_SERVICE_REC_LOCK (mapi_store, connect_lock);
 	exchange_mapi_connection_fetch_item (id_folder, id_message, 
 					camel_GetPropsList, G_N_ELEMENTS (camel_GetPropsList), 
 					camel_build_name_id, NULL, 
 					fetch_item_cb, &item, 
 					options);
+	CAMEL_SERVICE_REC_UNLOCK (mapi_store, connect_lock);
 
 	if (item == NULL) {
 		camel_exception_set (ex, CAMEL_EXCEPTION_SERVICE_INVALID, _("Could not get message"));
@@ -1809,6 +1831,8 @@ mapi_get_message_info(CamelFolder *folder, const char *uid)
 		mi = (CamelMessageInfoBase *)msg_info ;
 		return (msg_info);
 	}
+	/* Go online and fetch message summary. */
+
 	msg_info = camel_message_info_new(folder->summary);
 	mi = (CamelMessageInfoBase *)msg_info ;
 
@@ -1952,12 +1976,12 @@ mapi_transfer_messages_to (CamelFolder *source, GPtrArray *uids,
 	folder_id =  camel_mapi_store_folder_id_lookup (mapi_store, source->full_name) ;
 	exchange_mapi_util_mapi_id_from_string (folder_id, &src_fid);
 
-	folder_id =  camel_mapi_store_folder_id_lookup (mapi_store, destination->full_name) ;
+	folder_id = camel_mapi_store_folder_id_lookup (mapi_store, destination->full_name);
 	exchange_mapi_util_mapi_id_from_string (folder_id, &dest_fid);
 
 	for (i=0; i < uids->len; i++) {
 		mapi_id_t *mid = g_new0 (mapi_id_t, 1); /* FIXME : */
-		if (!exchange_mapi_util_mapi_ids_from_uid (g_ptr_array_index (uids, i), &src_fid, mid)) 
+		if (!exchange_mapi_util_mapi_ids_from_uid (g_ptr_array_index (uids, i), &src_fid, mid))
 			continue;
 
 		src_msg_ids = g_slist_prepend (src_msg_ids, mid);
@@ -2017,13 +2041,17 @@ mapi_append_message (CamelFolder *folder, CamelMimeMessage *message,
 	MapiItem *item = NULL;
 	mapi_id_t fid = 0, mid = 0;
 	const gchar *folder_id;
+	guint32 folder_flags = 0;
 
 	/*Reject outbox / sent & trash*/
-	si = camel_store_summary_path ((CamelStoreSummary *)mapi_store->summary,
-				       folder->full_name);
+	si = camel_store_summary_path ((CamelStoreSummary *)mapi_store->summary, folder->full_name);
+	if (si) {
+		folder_flags = si->flags;
+		camel_store_summary_info_free ((CamelStoreSummary *)mapi_store->summary, si);
+	}
 
-	if (((si->flags & CAMEL_FOLDER_TYPE_MASK) == CAMEL_FOLDER_TYPE_TRASH) ||
-	    ((si->flags & CAMEL_FOLDER_TYPE_MASK) == CAMEL_FOLDER_TYPE_OUTBOX)) {
+	if (((folder_flags & CAMEL_FOLDER_TYPE_MASK) == CAMEL_FOLDER_TYPE_TRASH) ||
+	    ((folder_flags & CAMEL_FOLDER_TYPE_MASK) == CAMEL_FOLDER_TYPE_OUTBOX)) {
 		camel_exception_setv (ex, CAMEL_EXCEPTION_SYSTEM, 
 				      _("Cannot append message to folder '%s'"),
 				      folder->full_name);
@@ -2202,5 +2230,3 @@ camel_mapi_folder_new (CamelStore *store, const char *folder_name, const char *f
 	}
 	return folder;
 }
-
-
